@@ -1,14 +1,20 @@
-import pytest
+""" 
+# hdl21 Unit Tests 
+"""
+
+import sys, copy, pytest
 from io import StringIO
 from types import SimpleNamespace
-from enum import Enum, auto
+from enum import Enum, EnumMeta, auto
+from textwrap import dedent
 
 # Import the PUT (package under test)
 import hdl21 as h
+import vlsir
 
 
 def test_version():
-    assert h.__version__ == "0.1.0"
+    assert h.__version__ == "0.2.0"
 
 
 def test_module1():
@@ -227,7 +233,7 @@ def test_params4():
     o2 = Outer(**d2)
     uname2 = h.params._unique_name(o2)
     assert uname1 == uname2
-    assert uname1 == "Outer(3dcc309796996b3a8a61db66631c5a93)"
+    assert uname1 == "3dcc309796996b3a8a61db66631c5a93"
 
 
 def test_bad_params1():
@@ -385,11 +391,11 @@ def test_cycle1():
 
     assert len(b2.instances) == 2
     assert len(b2.instarrays) == 0
-    assert len(b2.interfaces) == 0
+    assert len(b2.bundles) == 0
     assert len(b2.ports) == 0
     assert len(b2.signals) == 2
-    assert "_t2_inp_t1_out_" in b2.signals
-    assert "_t2_out_t1_inp_" in b2.signals
+    assert "t2_inp_t1_out" in b2.signals
+    assert "t2_out_t1_inp" in b2.signals
 
 
 def test_gen3():
@@ -476,22 +482,22 @@ def test_prim_proto1():
 
     ppkg = h.to_proto(HasPrims)
 
-    assert isinstance(ppkg, h.proto.Package)
+    assert isinstance(ppkg, vlsir.circuit.Package)
     assert len(ppkg.modules) == 1
     assert ppkg.domain == ""
 
     # Check the proto-Module
     pm = ppkg.modules[0]
-    assert isinstance(pm, h.proto.Module)
+    assert isinstance(pm, vlsir.circuit.Module)
     assert pm.name == "hdl21.tests.test_hdl21.HasPrims"
     assert len(pm.ports) == 0
     assert len(pm.signals) == 2
     assert len(pm.instances) == 5
-    assert len(pm.default_parameters) == 0
+    assert len(pm.parameters) == 0
     for inst in pm.instances:
-        assert isinstance(inst, h.proto.Instance)
+        assert isinstance(inst, vlsir.circuit.Instance)
         assert inst.module.WhichOneof("to") == "external"
-        assert inst.module.external.domain in ["hdl21.ideal", "hdl21.primitives"]
+        assert inst.module.external.domain in ["vlsir.primitives", "hdl21.primitives"]
 
     ns = h.from_proto(ppkg)
 
@@ -508,94 +514,109 @@ def test_prim_proto1():
         assert isinstance(inst._resolved, h.primitives.PrimitiveCall)
 
 
-def test_intf1():
-    # Create an interface
+def test_bundle1():
+    # Create an bundle
 
-    i1 = h.Interface(name="MyFirstInterface")
+    i1 = h.Bundle(name="MyFirstBundle")
     i1.s1 = h.Signal()
     i1.s2 = h.Signal()
 
-    assert isinstance(i1, h.Interface)
+    assert isinstance(i1, h.Bundle)
     assert isinstance(i1.s1, h.Signal)
     assert isinstance(i1.s2, h.Signal)
 
     ii1 = i1()
-    assert isinstance(ii1, h.InterfaceInstance)
+    assert isinstance(ii1, h.BundleInstance)
     assert ii1.role is None
     assert ii1.port == False
 
 
-def test_intf2():
-    # Wire up a few Modules via interfaces
+def test_bundle2():
+    # Wire up a few Modules via bundles
 
-    MySecondInterface = h.Interface(name="MySecondInterface")
-    MySecondInterface.s = h.Signal()
+    MySecondBundle = h.Bundle(name="MySecondBundle")
+    MySecondBundle.s = h.Signal()
 
     m1 = h.Module(name="M1")
-    m1.i = MySecondInterface(port=True)
+    m1.i = MySecondBundle(port=True)
     m2 = h.Module(name="M2")
-    m2.i = MySecondInterface(port=True)
+    m2.i = MySecondBundle(port=True)
 
     # Now create a parent Module connecting the two
     m3 = h.Module(name="M3")
     m3.i1 = m1()
     m3.i2 = m2(i=m3.i1.i)
-    assert "_i2_i_i1_i_" not in m3.namespace
+    assert "i2_i_i1_i" not in m3.namespace
 
-    # First run the "implicit interfaces" pass, and see that an explicit one is created
+    # First run the "implicit bundles" pass, and see that an explicit one is created
     from hdl21.elab import ElabPass
 
-    m3 = h.elaborate(m3, passes=[ElabPass.IMPLICIT_INTERFACES])
+    m3 = h.elaborate(m3, passes=[ElabPass.IMPLICIT_BUNDLES])
     assert isinstance(m3, h.Module)
     assert isinstance(m3.i1, h.Instance)
     assert isinstance(m3.i2, h.Instance)
-    assert "_i2_i_i1_i_" in m3.namespace
+    assert "i2_i_i1_i" in m3.namespace
 
     # Now elaborate it the rest of the way, to scalar signals
     m3 = h.elaborate(m3)
-    assert "_i2_i_i1_i_" not in m3.namespace
-    assert "__i2_i_i1_i__s_" in m3.namespace
-    assert isinstance(m3.get("__i2_i_i1_i__s_"), h.Signal)
-    assert m3.get("__i2_i_i1_i__s_") in m3.i1.conns.values()
+    assert "i2_i_i1_i" not in m3.namespace
+    assert "i2_i_i1_i_s" in m3.namespace
+    assert isinstance(m3.get("i2_i_i1_i_s"), h.Signal)
+    assert m3.get("i2_i_i1_i_s") in m3.i1.conns.values()
 
 
-def test_intf3():
-    # Test the interface-definition decorator
+def test_bundle3():
+    # Test the bundle-definition decorator
 
-    @h.interface
-    class Diff:  # Differential Signal Interface
+    @h.bundle
+    class Diff:  # Differential Signal Bundle
         p = h.Signal()
         n = h.Signal()
 
-    @h.interface
+    @h.bundle
     class DisplayPort:  # DisplayPort, kinda
         main_link = Diff()
         aux = h.Signal()
 
-    assert isinstance(DisplayPort, h.Interface)
-    assert isinstance(DisplayPort(), h.InterfaceInstance)
-    assert isinstance(DisplayPort.main_link, h.InterfaceInstance)
+    assert isinstance(DisplayPort, h.Bundle)
+    assert isinstance(DisplayPort(), h.BundleInstance)
+    assert isinstance(DisplayPort.main_link, h.BundleInstance)
     assert isinstance(DisplayPort.main_link.p, h.PortRef)
     assert isinstance(DisplayPort.main_link.n, h.PortRef)
     assert isinstance(DisplayPort.aux, h.Signal)
-    assert isinstance(Diff, h.Interface)
-    assert isinstance(Diff(), h.InterfaceInstance)
+    assert isinstance(Diff, h.Bundle)
+    assert isinstance(Diff(), h.BundleInstance)
     assert isinstance(Diff.p, h.Signal)
     assert isinstance(Diff.n, h.Signal)
 
+    # Instantiate one in a Module
+    m = h.Module(name="M")
+    m.dp = DisplayPort()
+    assert isinstance(m.dp, h.BundleInstance)
+    assert len(m.bundles) == 1
+    assert len(m.signals) == 0
 
-def test_intf4():
-    # Test interface roles
+    # And elaborate it
+    h.elaborate(m)
 
-    from enum import Enum, EnumMeta, auto
+    assert not hasattr(m, "dp")
+    assert len(m.bundles) == 0
+    assert len(m.signals) == 3
+    assert isinstance(m.get("dp_aux"), h.Signal)
+    assert isinstance(m.get("dp_main_link_p"), h.Signal)
+    assert isinstance(m.get("dp_main_link_n"), h.Signal)
 
-    @h.interface
-    class Diff:  # Differential Signal Interface
+
+def test_bundle4():
+    # Test bundle roles
+
+    @h.bundle
+    class Diff:  # Differential Signal Bundle
         p = h.Signal()
         n = h.Signal()
 
-    @h.interface
-    class HasRoles:  # An Interface with Roles
+    @h.bundle
+    class HasRoles:  # An Bundle with Roles
         class Roles(Enum):  # USB-Style Role Nomenclature
             HOST = auto()
             DEVICE = auto()
@@ -609,52 +630,52 @@ def test_intf4():
         rxd = Diff(src=Roles.DEVICE, dest=Roles.HOST)
 
     hr = HasRoles()
-    assert isinstance(HasRoles, h.Interface)
+    assert isinstance(HasRoles, h.Bundle)
     assert isinstance(HasRoles.roles, EnumMeta)
     assert isinstance(HasRoles.Roles, EnumMeta)
-    assert isinstance(hr, h.InterfaceInstance)
+    assert isinstance(hr, h.BundleInstance)
     assert isinstance(HasRoles.tx, h.Signal)
     assert isinstance(HasRoles.rx, h.Signal)
-    assert isinstance(HasRoles.txd, h.InterfaceInstance)
-    assert isinstance(HasRoles.rxd, h.InterfaceInstance)
+    assert isinstance(HasRoles.txd, h.BundleInstance)
+    assert isinstance(HasRoles.rxd, h.BundleInstance)
 
     @h.module
     class Host:
-        # A thing with a HOST-roled interface-port
-        port_ = HasRoles(port=True, role=HasRoles.Roles.HOST)
+        # A thing with a HOST-roled bundle-port
+        hr = HasRoles(port=True, role=HasRoles.Roles.HOST)
 
     @h.module
     class Device:
-        # A thing with a DEVICE-roled interface-port
-        port_ = HasRoles(port=True, role=HasRoles.Roles.DEVICE)
+        # A thing with a DEVICE-roled bundle-port
+        hr = HasRoles(port=True, role=HasRoles.Roles.DEVICE)
 
     @h.module
     class System:
         # Parent system-module including a host and device
         host = Host()
-        dev_ = Device(port_=host.port_)
+        devc = Device(hr=host.hr)
 
     assert isinstance(System, h.Module)
     assert isinstance(System.host, h.Instance)
-    assert isinstance(System.dev_, h.Instance)
-    assert "_dev__port__host_port__" not in System.namespace
+    assert isinstance(System.devc, h.Instance)
+    assert "devc_hr_host_hr" not in System.namespace
 
-    # First run the "implicit interfaces" pass, and see that an explicit one is created
+    # First run the "implicit bundles" pass, and see that an explicit one is created
     from hdl21.elab import ElabPass
 
-    sys = h.elaborate(System, passes=[ElabPass.IMPLICIT_INTERFACES])
-    assert "_dev__port__host_port__" in sys.namespace
+    sys = h.elaborate(System, passes=[ElabPass.IMPLICIT_BUNDLES])
+    assert "devc_hr_host_hr" in sys.namespace
 
     # Now expand the rest of the way, down to scalar signals
-    # Check that interface went away, and its constituent signals replaced it
+    # Check that bundle went away, and its constituent signals replaced it
     sys = h.elaborate(sys)
-    assert "_dev__port__host_port__" not in sys.namespace
-    assert "__dev__port__host_port___tx_" in sys.namespace
-    assert "__dev__port__host_port___rx_" in sys.namespace
-    assert "__dev__port__host_port____txd_p__" in sys.namespace
-    assert "__dev__port__host_port____txd_n__" in sys.namespace
-    assert "__dev__port__host_port____rxd_p__" in sys.namespace
-    assert "__dev__port__host_port____rxd_n__" in sys.namespace
+    assert "devc_hr_host_hr" not in sys.namespace
+    assert "devc_hr_host_hr_tx" in sys.namespace
+    assert "devc_hr_host_hr_rx" in sys.namespace
+    assert "devc_hr_host_hr_txd_p" in sys.namespace
+    assert "devc_hr_host_hr_txd_n" in sys.namespace
+    assert "devc_hr_host_hr_rxd_p" in sys.namespace
+    assert "devc_hr_host_hr_rxd_n" in sys.namespace
 
 
 def test_proto1():
@@ -662,15 +683,15 @@ def test_proto1():
 
     m = h.Module(name="TestProto1")
     ppkg = h.to_proto(m)
-    assert isinstance(ppkg, h.proto.Package)
+    assert isinstance(ppkg, vlsir.circuit.Package)
     assert len(ppkg.modules) == 1
     assert ppkg.domain == ""
     pm = ppkg.modules[0]
-    assert isinstance(pm, h.proto.Module)
+    assert isinstance(pm, vlsir.circuit.Module)
     assert pm.name == "hdl21.tests.test_hdl21.TestProto1"
     assert len(pm.ports) == 0
     assert len(pm.instances) == 0
-    assert len(pm.default_parameters) == 0
+    assert len(pm.parameters) == 0
 
 
 def test_proto2():
@@ -689,35 +710,35 @@ def test_proto2():
 
     ppkg = h.to_proto(TestProto2)
 
-    assert isinstance(ppkg, h.proto.Package)
+    assert isinstance(ppkg, vlsir.circuit.Package)
     assert len(ppkg.modules) == 3
     assert ppkg.domain == ""
 
     # Check the first Module in, Child1
     pm = ppkg.modules[0]
-    assert isinstance(pm, h.proto.Module)
+    assert isinstance(pm, vlsir.circuit.Module)
     assert pm.name == "hdl21.tests.test_hdl21.Child1"
     assert len(pm.ports) == 2
     assert len(pm.signals) == 0
     assert len(pm.instances) == 0
-    assert len(pm.default_parameters) == 0
+    assert len(pm.parameters) == 0
 
     # Check the second Module in, Child2
     pm = ppkg.modules[1]
-    assert isinstance(pm, h.proto.Module)
+    assert isinstance(pm, vlsir.circuit.Module)
     assert pm.name == "hdl21.tests.test_hdl21.Child2"
     assert len(pm.ports) == 2
     assert len(pm.signals) == 0
     assert len(pm.instances) == 0
-    assert len(pm.default_parameters) == 0
+    assert len(pm.parameters) == 0
 
     # And check the parent module
     pm = ppkg.modules[2]
-    assert isinstance(pm, h.proto.Module)
+    assert isinstance(pm, vlsir.circuit.Module)
     assert pm.name == "hdl21.tests.test_hdl21.TestProto2"
     assert len(pm.ports) == 0
     assert len(pm.instances) == 2
-    assert len(pm.default_parameters) == 0
+    assert len(pm.parameters) == 0
 
 
 def test_proto3():
@@ -731,31 +752,31 @@ def test_proto3():
     M2.s = h.Signal(width=4)
     M2.i = M1()
     M2.i(p1=M2.s[0])
-    M2.i(p8=h.Concat(M2.s, h.Concat(M2.s[0], M2.s[1]), M2.s[2:3]))
+    M2.i(p8=h.Concat(M2.s, h.Concat(M2.s[0], M2.s[1]), M2.s[2:4]))
 
     ppkg = h.to_proto(M2, domain="test_proto3")
 
-    assert isinstance(ppkg, h.proto.Package)
+    assert isinstance(ppkg, vlsir.circuit.Package)
     assert len(ppkg.modules) == 2
     assert ppkg.domain == "test_proto3"
 
     # Check the child module
     pm = ppkg.modules[0]
-    assert isinstance(pm, h.proto.Module)
+    assert isinstance(pm, vlsir.circuit.Module)
     assert pm.name == "hdl21.tests.test_hdl21.M1"
     assert len(pm.ports) == 2
     assert len(pm.signals) == 0
     assert len(pm.instances) == 0
-    assert len(pm.default_parameters) == 0
+    assert len(pm.parameters) == 0
 
     # And check the parent module
     pm = ppkg.modules[1]
-    assert isinstance(pm, h.proto.Module)
+    assert isinstance(pm, vlsir.circuit.Module)
     assert pm.name == "hdl21.tests.test_hdl21.M2"
     assert len(pm.ports) == 0
     assert len(pm.signals) == 1
     assert len(pm.instances) == 1
-    assert len(pm.default_parameters) == 0
+    assert len(pm.parameters) == 0
 
     ns = h.from_proto(ppkg)
 
@@ -776,13 +797,16 @@ def test_proto3():
     inst = M2.i
     assert isinstance(inst.conns["p1"], h.signal.Slice)
     assert inst.conns["p1"].signal is M2.s
-    assert inst.conns["p1"].top == 0
+    assert inst.conns["p1"].top == 1
     assert inst.conns["p1"].bot == 0
+    assert inst.conns["p1"].width == 1
     assert isinstance(inst.conns["p8"], h.signal.Concat)
-    assert len(inst.conns["p8"].parts) == 3
+    assert len(inst.conns["p8"].parts) == 4
     assert inst.conns["p8"].parts[0] is M2.s
-    assert isinstance(inst.conns["p8"].parts[1], h.signal.Concat)
+    assert isinstance(inst.conns["p8"].parts[0], h.signal.Signal)
+    assert isinstance(inst.conns["p8"].parts[1], h.signal.Slice)
     assert isinstance(inst.conns["p8"].parts[2], h.signal.Slice)
+    assert isinstance(inst.conns["p8"].parts[3], h.signal.Slice)
 
 
 def test_proto_roundtrip():
@@ -858,111 +882,167 @@ def test_proto_roundtrip2():
     assert len(M2.ports) == 3
     assert len(M2.signals) == 2
     assert "s" in M2.signals
-    assert "_i1_i_i0_o_" in M2.signals
+    assert "i1_i_i0_o" in M2.signals
 
 
-def test_bigger_interfaces():
-    # Test a slightly more elaborate Interface-based system
+def test_bigger_bundles():
+    """ Test a slightly more elaborate Bundle-based system """
 
-    class MsRoles(Enum):
-        MS = auto()
-        SL = auto()
+    class HostDevice(Enum):
+        HOST = auto()
+        DEVICE = auto()
 
-    @h.interface
+    @h.bundle
     class Jtag:
-        roles = MsRoles
-        tck = h.Signal(src=roles.MS, dest=roles.SL)
-        tdi = h.Signal(src=roles.MS, dest=roles.SL)
-        tms = h.Signal(src=roles.MS, dest=roles.SL)
-        tdo = h.Signal(src=roles.SL, dest=roles.MS)
+        # Jtag Bundle
 
-    @h.interface
+        roles = HostDevice
+        tck, tdi, tms = h.Signals(3, src=roles.HOST, dest=roles.DEVICE)
+        tdo = h.Signal(src=roles.DEVICE, dest=roles.HOST)
+
+    @h.bundle
+    class Uart:
+        # Uart Bundle
+        class Roles(Enum):
+            # Uart roles are essentially peers, here named `ME` and `YOU`.
+            # Essentially everything will use the role `ME`,
+            # except for interconnect which swaps between the two.
+            ME = auto()
+            YOU = auto()
+
+        tx = h.Signal(src=Roles.ME, dest=Roles.YOU)
+        rx = h.Signal(src=Roles.YOU, dest=Roles.ME)
+
+    @h.bundle
     class Spi:
-        roles = MsRoles
-        sck = h.Signal(src=roles.MS, dest=roles.SL)
-        cs = h.Signal(src=roles.MS, dest=roles.SL)
-        dq = h.Signal(src=roles.SL, dest=roles.MS, width=4)
+        # Spi Bundle
+        roles = HostDevice
+        sck, cs = h.Signals(2, src=roles.HOST, dest=roles.DEVICE)
+        dq = h.Signal(src=roles.DEVICE, dest=roles.HOST, width=4)
 
     @h.module
     class Chip:
-        spi = Spi(role=MsRoles.MS, port=True)
-        jtag = Jtag(role=MsRoles.SL, port=True)
+        spi = Spi(role=HostDevice.HOST, port=True)
+        jtag = Jtag(role=HostDevice.DEVICE, port=True)
+        uart = Uart(role=Uart.Roles.ME, port=True)
         ...  # Actual internal content, which likely connects these down *many* levels of hierarchy
 
     @h.module
     class SpiFlash:
         # A typical flash memory with a SPI port
-        spi = Spi(role=MsRoles.SL, port=True)
+        spi = Spi(role=HostDevice.DEVICE, port=True)
 
     @h.module
     class Board:
         # A typical embedded board, featuring a custom chip, SPI-connected flash, and JTAG port
-        jtag = Jtag(role=MsRoles.SL, port=True)
-        chip = Chip(jtag=jtag)
+        jtag = Jtag(role=HostDevice.DEVICE, port=True)
+        uart = Uart(role=Uart.Roles.ME, port=True)
+
+        chip = Chip(jtag=jtag, uart=uart)
         flash = SpiFlash(spi=chip.spi)
 
     @h.module
     class Tester:
         # A typical test-widget with a JTAG port
-        jtag = Jtag(role=MsRoles.MS, port=True)
+        jtag = Jtag(role=HostDevice.HOST, port=True)
+        uart = Uart(role=Uart.Roles.ME, port=True)
 
     @h.module
     class TestSystem:
         # A system in which `Tester` can test `Board`
         jtag = Jtag()
+
         tester = Tester(jtag=jtag)
         board = Board(jtag=jtag)
 
+        # Connect UART, swapping `rx` and `tx`
+        u0, u1 = h.Signals(2)
+        board.uart = h.AnonymousBundle(tx=u0, rx=u1)
+        tester.uart = h.AnonymousBundle(rx=u0, tx=u1)
+
+    assert isinstance(TestSystem.jtag, h.BundleInstance)
+    assert isinstance(TestSystem.tester, h.Instance)
+    assert isinstance(TestSystem.board, h.Instance)
+
+    assert isinstance(TestSystem.tester.uart, h.PortRef)
+    assert isinstance(TestSystem.tester.uart.tx, h.PortRef)
+    assert isinstance(TestSystem.tester.uart.rx, h.PortRef)
+    assert isinstance(TestSystem.board.uart, h.PortRef)
+    assert isinstance(TestSystem.board.uart.tx, h.PortRef)
+    assert isinstance(TestSystem.board.uart.rx, h.PortRef)
+
     # Run this through elaboration
     h.elaborate(TestSystem)
 
     # Post-elab checks
-    assert not hasattr(TestSystem, "jtag")
-    assert len(TestSystem.ports) == 0
-    assert len(TestSystem.signals) == 4
-    assert len(TestSystem.instances) == 2
+    # TestSystem
+    assert isinstance(TestSystem.tester, h.Instance)
+    assert isinstance(TestSystem.board, h.Instance)
     assert TestSystem.tester.of is Tester
     assert TestSystem.board.of is Board
-    assert isinstance(TestSystem.get("_jtag_tck_"), h.Signal)
-    assert isinstance(TestSystem.get("_jtag_tdi_"), h.Signal)
-    assert isinstance(TestSystem.get("_jtag_tdo_"), h.Signal)
-    assert isinstance(TestSystem.get("_jtag_tms_"), h.Signal)
-    assert isinstance(Tester.get("_jtag_tck_"), h.Signal)
-    assert len(Tester.ports) == 4
+    assert not hasattr(TestSystem, "jtag")
+    assert not hasattr(TestSystem, "uart")
+    assert "u0" in TestSystem.namespace
+    assert "u1" in TestSystem.namespace
+    assert len(TestSystem.ports) == 0
+    assert len(TestSystem.signals) == 6
+    assert len(TestSystem.instances) == 2
+    assert isinstance(TestSystem.get("u0"), h.Signal)
+    assert isinstance(TestSystem.get("u1"), h.Signal)
+    assert isinstance(TestSystem.get("jtag_tck"), h.Signal)
+    assert isinstance(TestSystem.get("jtag_tdi"), h.Signal)
+    assert isinstance(TestSystem.get("jtag_tdo"), h.Signal)
+    assert isinstance(TestSystem.get("jtag_tms"), h.Signal)
+
+    # Tester
+    assert len(Tester.ports) == 6
     assert len(Tester.signals) == 0
     assert len(Tester.instances) == 0
-    assert Tester.get("_jtag_tck_").vis == h.signal.Visibility.PORT
-    assert Tester.get("_jtag_tdo_").vis == h.signal.Visibility.PORT
-    assert Tester.get("_jtag_tdi_").vis == h.signal.Visibility.PORT
-    assert Tester.get("_jtag_tms_").vis == h.signal.Visibility.PORT
-    assert len(Board.ports) == 4
+    assert isinstance(Tester.get("jtag_tck"), h.Signal)
+    assert Tester.get("jtag_tck").vis == h.signal.Visibility.PORT
+    assert Tester.get("jtag_tdo").vis == h.signal.Visibility.PORT
+    assert Tester.get("jtag_tdi").vis == h.signal.Visibility.PORT
+    assert Tester.get("jtag_tms").vis == h.signal.Visibility.PORT
+    assert Tester.get("uart_tx").vis == h.signal.Visibility.PORT
+    assert Tester.get("uart_rx").vis == h.signal.Visibility.PORT
+
+    # Board
+    assert len(Board.ports) == 6
     assert len(Board.signals) == 3  # SPI signals
     assert len(Board.instances) == 2
     assert Board.chip.of is Chip
     assert Board.flash.of is SpiFlash
-    assert Board.get("_jtag_tck_").vis == h.signal.Visibility.PORT
-    assert Board.get("_jtag_tdo_").vis == h.signal.Visibility.PORT
-    assert Board.get("_jtag_tdi_").vis == h.signal.Visibility.PORT
-    assert Board.get("_jtag_tms_").vis == h.signal.Visibility.PORT
-    assert Board.get("__flash_spi_chip_spi__sck_").vis == h.signal.Visibility.INTERNAL
-    assert Board.get("__flash_spi_chip_spi__cs_").vis == h.signal.Visibility.INTERNAL
-    assert Board.get("__flash_spi_chip_spi__dq_").vis == h.signal.Visibility.INTERNAL
-    assert len(Chip.ports) == 7
+    assert Board.get("jtag_tck").vis == h.signal.Visibility.PORT
+    assert Board.get("jtag_tdo").vis == h.signal.Visibility.PORT
+    assert Board.get("jtag_tdi").vis == h.signal.Visibility.PORT
+    assert Board.get("jtag_tms").vis == h.signal.Visibility.PORT
+    assert Board.get("uart_tx").vis == h.signal.Visibility.PORT
+    assert Board.get("uart_rx").vis == h.signal.Visibility.PORT
+    assert Board.get("flash_spi_chip_spi_sck").vis == h.signal.Visibility.INTERNAL
+    assert Board.get("flash_spi_chip_spi_cs").vis == h.signal.Visibility.INTERNAL
+    assert Board.get("flash_spi_chip_spi_dq").vis == h.signal.Visibility.INTERNAL
+
+    # Chip
+    assert len(Chip.ports) == 9
     assert len(Chip.signals) == 0
     assert len(Chip.instances) == 0
-    assert Chip.get("_jtag_tck_").vis == h.signal.Visibility.PORT
-    assert Chip.get("_jtag_tdo_").vis == h.signal.Visibility.PORT
-    assert Chip.get("_jtag_tdi_").vis == h.signal.Visibility.PORT
-    assert Chip.get("_jtag_tms_").vis == h.signal.Visibility.PORT
-    assert Chip.get("_spi_sck_").vis == h.signal.Visibility.PORT
-    assert Chip.get("_spi_cs_").vis == h.signal.Visibility.PORT
-    assert Chip.get("_spi_dq_").vis == h.signal.Visibility.PORT
+    assert Chip.get("jtag_tck").vis == h.signal.Visibility.PORT
+    assert Chip.get("jtag_tdo").vis == h.signal.Visibility.PORT
+    assert Chip.get("jtag_tdi").vis == h.signal.Visibility.PORT
+    assert Chip.get("jtag_tms").vis == h.signal.Visibility.PORT
+    assert Chip.get("spi_sck").vis == h.signal.Visibility.PORT
+    assert Chip.get("spi_cs").vis == h.signal.Visibility.PORT
+    assert Chip.get("spi_dq").vis == h.signal.Visibility.PORT
+    assert Chip.get("uart_tx").vis == h.signal.Visibility.PORT
+    assert Chip.get("uart_rx").vis == h.signal.Visibility.PORT
+
+    # SpiFlash
     assert len(SpiFlash.ports) == 3
     assert len(SpiFlash.signals) == 0
     assert len(SpiFlash.instances) == 0
-    assert SpiFlash.get("_spi_sck_").vis == h.signal.Visibility.PORT
-    assert SpiFlash.get("_spi_cs_").vis == h.signal.Visibility.PORT
-    assert SpiFlash.get("_spi_dq_").vis == h.signal.Visibility.PORT
+    assert SpiFlash.get("spi_sck").vis == h.signal.Visibility.PORT
+    assert SpiFlash.get("spi_cs").vis == h.signal.Visibility.PORT
+    assert SpiFlash.get("spi_dq").vis == h.signal.Visibility.PORT
 
 
 def test_signal_slice1():
@@ -971,42 +1051,47 @@ def test_signal_slice1():
     sig = h.Signal(width=10)
     sl = sig[0]
     assert isinstance(sl, h.signal.Slice)
-    assert sl.top == 0
+    assert sl.top == 1
     assert sl.bot == 0
     assert sl.width == 1
-    assert sl.step == 1
+    assert sl.step is None
     assert sl.signal is sig
 
-    sl = sig[0:4]
+    sl = sig[0:5]
     assert isinstance(sl, h.signal.Slice)
-    assert sl.top == 4
+    assert sl.top == 5
     assert sl.bot == 0
     assert sl.width == 5
+    assert sl.step is None
     assert sl.signal is sig
 
     sl = sig[:]
     assert isinstance(sl, h.signal.Slice)
-    assert sl.top == 9
+    assert sl.top == 10
     assert sl.bot == 0
     assert sl.width == 10
+    assert sl.step is None
     assert sl.signal is sig
 
 
 def test_signal_slice2():
     # Test slicing advanced features
     sl = h.Signal(width=11)[-1]
-    assert sl.top == 10
+    assert sl.top == 11
     assert sl.bot == 10
+    assert sl.step is None
     assert sl.width == 1
 
     sl = h.Signal(width=11)[:-1]
-    assert sl.top == 9
+    assert sl.top == 10
     assert sl.bot == 0
+    assert sl.step is None
     assert sl.width == 10
 
     sl = h.Signal(width=11)[-2:]
-    assert sl.top == 10
+    assert sl.top == 11
     assert sl.bot == 9
+    assert sl.step is None
     assert sl.width == 2
 
 
@@ -1122,7 +1207,7 @@ def test_generator_recall():
     nl = StringIO()
     h.netlist(ppkg, nl)
     nl = nl.getvalue()
-    assert "CallMeTwice_NoParams___" in nl
+    assert "CallMeTwice_NoParams" in nl
     assert "Caller" in nl
 
     # Round-trip back from Proto to Modules
@@ -1172,10 +1257,10 @@ def test_mos_generator():
     assert m.instances["unit0"].of.params == h.primitives.MosParams()
     assert m.instances["unit1"].of.params == h.primitives.MosParams()
     assert len(m.signals) == 1
-    assert "_unit1_s_unit0_d_" in m.signals
+    assert "unit1_s_unit0_d" in m.signals
 
     ppkg = h.to_proto(m)
-    assert isinstance(ppkg, h.proto.Package)
+    assert isinstance(ppkg, vlsir.circuit.Package)
 
 
 def test_series_parallel_generator():
@@ -1207,11 +1292,11 @@ def test_series_parallel_generator():
     for inst in m.instances.values():
         assert inst.of is M
     assert len(m.signals) == 2
-    assert "_unit_0_1_a_unit_0_0_b_" in m.signals
-    assert "_unit_1_1_a_unit_1_0_b_" in m.signals
+    assert "unit_0_1_a_unit_0_0_b" in m.signals
+    assert "unit_1_1_a_unit_1_0_b" in m.signals
 
     ppkg = h.to_proto(m)
-    assert isinstance(ppkg, h.proto.Package)
+    assert isinstance(ppkg, vlsir.circuit.Package)
 
 
 def test_instance_mult():
@@ -1276,12 +1361,12 @@ def test_instance_mult2():
     # Check that array-flattening completed correctly
     assert len(Parent.instances) == 3
     assert len(Parent.instarrays) == 0
-    assert Parent.get("_child_0_").of is Child
-    assert Parent.get("_child_1_").of is Child
-    assert Parent.get("_child_2_").of is Child
-    assert Parent.get("_child_0_").conns["p"] == Parent.a[0]
-    assert Parent.get("_child_1_").conns["p"] == Parent.a[1]
-    assert Parent.get("_child_2_").conns["p"] == Parent.a[2]
+    assert Parent.get("child_0").of is Child
+    assert Parent.get("child_1").of is Child
+    assert Parent.get("child_2").of is Child
+    assert Parent.get("child_0").conns["p"] == Parent.a[0]
+    assert Parent.get("child_1").conns["p"] == Parent.a[1]
+    assert Parent.get("child_2").conns["p"] == Parent.a[2]
 
 
 def test_instance_mult3():
@@ -1301,12 +1386,12 @@ def test_instance_mult3():
     # Check that array-flattening completed correctly
     assert len(Parent.instances) == 3
     assert len(Parent.instarrays) == 0
-    assert Parent.get("_child_0_").of is Child
-    assert Parent.get("_child_1_").of is Child
-    assert Parent.get("_child_2_").of is Child
-    assert Parent.get("_child_0_").conns["p"] == Parent.a[0]
-    assert Parent.get("_child_1_").conns["p"] == Parent.a[1]
-    assert Parent.get("_child_2_").conns["p"] == Parent.a[2]
+    assert Parent.get("child_0").of is Child
+    assert Parent.get("child_1").of is Child
+    assert Parent.get("child_2").of is Child
+    assert Parent.get("child_0").conns["p"] == Parent.a[0]
+    assert Parent.get("child_1").conns["p"] == Parent.a[1]
+    assert Parent.get("child_2").conns["p"] == Parent.a[2]
 
 
 def test_instance_mult4():
@@ -1328,15 +1413,15 @@ def test_instance_mult4():
     # Check that array-flattening completed correctly
     assert len(Parent.instances) == 3
     assert len(Parent.instarrays) == 0
-    assert Parent.get("_child_0_").of is Child
-    assert Parent.get("_child_1_").of is Child
-    assert Parent.get("_child_2_").of is Child
-    assert Parent.get("_child_0_").conns["a"] == Parent.a
-    assert Parent.get("_child_1_").conns["a"] == Parent.a
-    assert Parent.get("_child_2_").conns["a"] == Parent.a
-    assert Parent.get("_child_0_").conns["b"] == Parent.b[0:15]
-    assert Parent.get("_child_1_").conns["b"] == Parent.b[16:31]
-    assert Parent.get("_child_2_").conns["b"] == Parent.b[32:47]
+    assert Parent.get("child_0").of is Child
+    assert Parent.get("child_1").of is Child
+    assert Parent.get("child_2").of is Child
+    assert Parent.get("child_0").conns["a"] == Parent.a
+    assert Parent.get("child_1").conns["a"] == Parent.a
+    assert Parent.get("child_2").conns["a"] == Parent.a
+    assert Parent.get("child_0").conns["b"] == Parent.b[0:16]
+    assert Parent.get("child_1").conns["b"] == Parent.b[16:32]
+    assert Parent.get("child_2").conns["b"] == Parent.b[32:48]
 
 
 def test_netlist_fmts():
@@ -1388,6 +1473,56 @@ def test_netlist_fmts():
     assert ".p(p)" in nl
     assert "endmodule // Top" in nl
 
+    nl = StringIO()
+    h.netlist(ppkg, nl, "spice")
+    nl = nl.getvalue()
+    assert ".SUBCKT Bot \n+ s_2 s_1 s_0 p" in nl
+    assert ".SUBCKT Top \n+ p" in nl
+    assert "xb \n+ s_2 s_1 s_0 p \n+ Bot" in nl
+
+
+def test_spice_netlister():
+    @h.module
+    class DUT:
+        a = h.Input(width=5)
+        b = h.Output(width=5)
+        res = h.IdealResistor(h.ResistorParams(r=10e3))(p=a[0], n=b[0])
+        cap = h.IdealCapacitor(h.IdealCapacitorParams(c=10e-12))(p=a[1], n=b[1])
+        ind = h.IdealInductor(h.IdealInductorParams(l=10e-9))(p=a[2], n=b[2])
+
+    ppkg = h.to_proto(DUT)
+    nl = StringIO()
+    h.netlist(ppkg, nl, "spice")
+    good = dedent(
+        """\
+        .SUBCKT DUT 
+        + a_4 a_3 a_2 a_1 a_0 b_4 b_3 b_2 b_1 b_0 
+        + * No parameters
+
+        rres 
+        + a_0 b_0 
+        + 10000.0 
+        + * No parameters
+
+
+        ccap 
+        + a_1 b_1 
+        + 1e-11 
+        + * No parameters
+
+
+        lind 
+        + a_2 b_2 
+        + 1e-08 
+        + * No parameters
+
+
+        .ENDS
+    """
+    )
+    nl = nl.getvalue()
+    assert good in nl
+
 
 def test_bad_width_conn():
     """ Test invalid connection-widths """
@@ -1401,22 +1536,274 @@ def test_bad_width_conn():
         h.elaborate(q)
 
 
-def test_bad_intf_conn():
-    """ Test invalid Interface connections """
+def test_bad_bundle_conn():
+    """ Test invalid Bundle connections """
 
-    @h.interface
+    @h.bundle
     class P:
         p = h.Signal(width=3)
 
-    @h.interface
+    @h.bundle
     class R:
         z = h.Signal(width=11)
 
-    c = h.Module(name="c")
-    c.p = P(port=True)  # `P`-type Interface
-    q = h.Module(name="q")
-    q.r = R()  # `R`-type Interface
-    q.c = c(p=q.r)  # <= Bad connection here
+    @h.module
+    class C:
+        p = P(port=True)  # `P`-type Bundle
+
+    @h.module
+    class Q:
+        r = R()  # `R`-type Bundle
+        c = C(p=r)  # <= Bad connection here
 
     with pytest.raises(RuntimeError):
-        h.elaborate(q)
+        h.elaborate(Q)
+
+
+def test_illegal_module_attrs():
+    """ Test attempting to add illegal attributes """
+    m = h.Module()
+    with pytest.raises(TypeError):
+        m.a = list()
+
+    with pytest.raises(TypeError):
+
+        @h.module
+        class M:
+            a = list()
+
+    @h.module
+    class C:
+        ...
+
+    with pytest.raises(TypeError):
+        C.b = TabError
+
+    # Use combinations of legal attributes
+
+    m = h.Module(name="m")
+    with pytest.raises(TypeError):
+        m.p = 5
+    with pytest.raises(TypeError):
+        m.z = dict(a=h.Signal())
+    with pytest.raises(TypeError):
+        m.q = [h.Port()]
+    with pytest.raises(TypeError):
+        m.p = (h.Input(), h.Output())
+
+
+def test_copy_signal():
+    """ Copying a Signal """
+    copy.copy(h.Signal())
+
+
+def test_copy_bundle_instance():
+    """ Copying a BundleInstance """
+    # This generally fails when run in the debugger, but seems alright stand-alone (?)
+    copy.copy(h.BundleInstance(name="inst", of=h.Bundle()))
+
+
+def test_bundle_destructure():
+    """ Test de-structuring bundles to individual Signals """
+
+    @h.bundle
+    class B:
+        w1 = h.Signal(width=1)
+        w3 = h.Signal(width=3)
+
+    @h.module
+    class Child:
+        w1 = h.Port(width=1)
+        w3 = h.Port(width=3)
+
+    @h.module
+    class Parent:
+        b = B()  # `B`-type Bundle instance
+        c = Child(w1=b.w1, w3=b.w3)  # `Child`-type instance, connected to `b`
+
+    h.elaborate(Parent)
+
+    assert len(Parent.instances) == 1
+    assert len(Parent.signals) == 2
+    assert "b_w1" in Parent.signals
+    assert "b_w3" in Parent.signals
+    assert Parent.c.conns["w1"] is Parent.signals["b_w1"]
+    assert Parent.c.conns["w3"] is Parent.signals["b_w3"]
+
+
+def test_orphanage():
+    """ Test that orphaned Module-attributes fail at elaboration """
+    m1 = h.Module(name="m1")
+    m1.s = h.Signal()  # Signal `s` is now "parented" by `m1`
+
+    m2 = h.Module(name="m2")
+    m2.y = m1.s  # Now `s` has been "orphaned" (or perhaps "cradle-robbed") by `m2`
+
+    # Note elaborating `m2` continues to work
+    h.elaborate(m2)
+
+    with pytest.raises(RuntimeError):
+        # Elaborating `m1` should fail, since `s` is orphaned
+        h.elaborate(m1)
+
+
+def test_orphanage2():
+    """ Test orphaning a bundle instance """
+
+    @h.bundle
+    class B:
+        bb = h.Signal(width=1)
+
+    b = B()  # Instance of `B`-type Bundle, to be orphaned
+
+    m1 = h.Module(name="m1")
+    m1.b = b
+    m2 = h.Module(name="m2")
+    m2.b = b
+
+    # Note elaborating `m2` continues to work
+    h.elaborate(m2)
+
+    with pytest.raises(RuntimeError):
+        # Elaborating `m1` should fail, since `s` is orphaned
+        h.elaborate(m1)
+
+
+def test_orphanage3():
+    """ Test orphaning an instance """
+    I = h.Module(name="I")
+    i = I()  # Instance to be orphaned
+
+    m1 = h.Module(name="m1")
+    m1.i = i
+    m2 = h.Module(name="m2")
+    m2.i = i
+
+    # Note elaborating `m2` continues to work
+    h.elaborate(m2)
+
+    with pytest.raises(RuntimeError):
+        # Elaborating `m1` should fail, since `s` is orphaned
+        h.elaborate(m1)
+
+
+@pytest.mark.xfail(reason="#6")
+def test_wrong_decorator():
+    """ Mistake `Module` for `module` """
+
+    with pytest.raises(RuntimeError):
+
+        @h.Module  # Bad!
+        class M:
+            ...
+
+
+def test_elab_noconn():
+    """ Initial test of elaborating a `NoConn` """
+
+    @h.module
+    class Inner:
+        p = h.Port()
+
+    @h.module
+    class HasNoConn:
+        i1 = Inner()
+        i1.p = h.NoConn()
+
+    h.elaborate(HasNoConn)
+
+    assert len(HasNoConn.signals) == 1
+
+
+def test_bad_noconn():
+    """ Test that a doubly-connected `NoConn` should fail """
+
+    @h.module
+    class Inner:
+        p = h.Port()
+
+    @h.module
+    class Bad:
+        # Create two instances of `Inner`
+        i1 = Inner()
+        i2 = Inner()
+
+        # Connect the first to a `NoConn`
+        i1.p = h.NoConn()
+        # And then connect the second to the first.
+        # Herein lies the "problem"
+        i2.p = i1.p
+
+    with pytest.raises(RuntimeError):
+        h.elaborate(Bad)
+
+
+def test_array_concat_conn():
+    """ Test connecting a `Concat` to an `InstArray` """
+
+    Child = h.Module(name="Child")
+    Child.p3 = h.Input(width=3)
+    Child.p5 = h.Input(width=5)
+
+    Parent = h.Module(name="Parent")
+    Parent.s5 = h.Signal(width=5)
+    Parent.s9 = h.Signal(width=9)
+    Parent.c = 2 * Child()
+    Parent.c.p3 = Parent.s5[0:3]
+    Parent.c.p5 = h.Concat(Parent.s5[3], Parent.s9)
+
+    h.netlist(h.to_proto(Parent), StringIO(), "verilog")
+
+
+def test_slice_resolution():
+    """ Test resolutions of slice combinations """
+
+    from hdl21.elab import _resolve_slice
+
+    # Slice of a Signal
+    s = h.Signal(width=5)
+    assert _resolve_slice(s[0]) == s[0]
+
+    # Slice of a Concat
+    s1 = h.Signal(name="s1", width=5)
+    s2 = h.Signal(name="s2", width=5)
+    c = h.Concat(s1, s2)
+    assert _resolve_slice(c[0]) == s1[0]
+    assert _resolve_slice(c[1]) == s1[1]
+    assert _resolve_slice(c[2]) == s1[2]
+    assert _resolve_slice(c[3]) == s1[3]
+    assert _resolve_slice(c[4]) == s1[4]
+    assert _resolve_slice(c[5]) == s2[0]
+    assert _resolve_slice(c[6]) == s2[1]
+    assert _resolve_slice(c[7]) == s2[2]
+    assert _resolve_slice(c[8]) == s2[3]
+    assert _resolve_slice(c[9]) == s2[4]
+
+    # Slice of a Slice
+    sa = h.Signal(name="sa", width=5)
+    assert _resolve_slice(sa[2:5][0]) == sa[2]
+    assert _resolve_slice(sa[2:5][1]) == sa[3]
+    assert _resolve_slice(sa[2:5][2]) == sa[4]
+
+    assert _resolve_slice(sa[:][0]) == sa[0]
+    assert _resolve_slice(sa[:][1]) == sa[1]
+    assert _resolve_slice(sa[:][2]) == sa[2]
+    assert _resolve_slice(sa[:][3]) == sa[3]
+    assert _resolve_slice(sa[:][4]) == sa[4]
+
+    # Check a concatenation case. Note `Concat` does not support equality, so compare parts.
+    assert _resolve_slice(sa[:][0:4]).parts == (sa[0], sa[1], sa[2], sa[3])
+
+
+@pytest.mark.xfail(reason="#1")
+def test_export_strides():
+    """ Test exporting connections with non-unit Slice-strides """
+
+    c = h.Module(name="c")
+    c.p = h.Input(width=2)
+
+    p = h.Module(name="p")
+    p.s = h.Signal(width=20)
+    p.c = c(p=p.s[::10])  # Connect two of the 20 bits, with stride 10
+
+    h.netlist(h.to_proto(p), sys.stdout, "verilog")
